@@ -4,7 +4,7 @@
 #
 # Author: Slaven Rezic
 #
-# Copyright (C) 2017,2018,2019,2020,2022,2023,2024 Slaven Rezic. All rights reserved.
+# Copyright (C) 2017,2018,2019,2020,2022,2023,2024,2025 Slaven Rezic. All rights reserved.
 # This program is free software; you can redistribute it and/or
 # modify it under the same terms as Perl itself.
 #
@@ -17,7 +17,7 @@ use warnings;
 
 {
     package Doit;
-    our $VERSION = '0.028_54';
+    our $VERSION = '0.028_55';
     $VERSION =~ s{_}{};
 
     use constant IS_WIN => $^O eq 'MSWin32';
@@ -1886,6 +1886,9 @@ use warnings;
     sub send_data {
 	my($self, @cmd) = @_;
 	my $fh = $self->{outfh};
+	if (!defined $fh) {
+	    warning "Doit::RPC::send_data: output filehandle does not exist (anymore)";
+	}
 	my $data = Storable::nfreeze(\@cmd);
 	print $fh pack("N", length($data)) . $data;
     }
@@ -1973,22 +1976,43 @@ use warnings;
 	      }, $class;
     }
 
+    sub _label_string {
+	my $self = shift;
+	(defined $self->{label} ? "in connection '$self->{label}' " : "");
+    }
+
     # Call for every command on client
     sub call_remote {
 	my($self, @args) = @_;
+	my $do_exit = @args == 1 && $args[0] eq 'exit';
+	return if !$self->{outfh}; # already exited
 	my $context = wantarray ? 'a' : 's'; # XXX more possible context (void...)?
 	$self->send_data($context, @args);
-	my($rettype, @ret) = $self->receive_data(@args);
+	my($rettype, @ret) = $self->receive_data;
 	if (defined $rettype && $rettype eq 'e') {
 	    die $ret[0];
 	} elsif (defined $rettype && $rettype eq 'r') {
+	    if ($ret[0] eq 'bye-bye' && $do_exit) {
+		$self->{outfh} = undef; # remember that we called exit for next time and DESTROY
+	    }
 	    if ($context eq 'a') {
 		return @ret;
 	    } else {
 		return $ret[0];
 	    }
 	} else {
-	    die "Unexpected return type " . (defined $self->{label} ? "in connection '$self->{label}' " : "") . (defined $rettype ? "'$rettype'" : "<undefined>") . " (should be 'e' or 'r')";
+	    Doit::Log::error("Unexpected return type " . $self->_label_string . (defined $rettype ? "'$rettype'" : "<undefined>") . " (should be 'e' or 'r')");
+	}
+    }
+
+    sub wait_ready {
+	my($self) = @_;
+	my $ret = $self->call_remote('__doit_rpc_ping');
+        if ($ret ne "pong") {
+	    Doit::Log::warning("Unexpected return value in wait_ready call " . $self->_label_string . "'$ret' (should be 'pong')");
+	    0;
+	} else {
+	    1;
 	}
     }
 }
@@ -2023,7 +2047,7 @@ use warnings;
 	my $d;
 	if ($self->{debug}) {
 	    $d = sub ($) {
-		Doit::Log::info("WORKER: $_[0]");
+		Doit::Log::info("WORKER: $_[0]\r");
 	    };
 	} else {
 	    $d = sub ($) { };
@@ -2056,11 +2080,15 @@ use warnings;
 		$d->(" got eof");
 		$fh->close;
 		return;
-	    } elsif ($data[0] =~ m{^exit$}) {
+	    } elsif ($data[0] eq 'exit') {
 		$d->(" got exit command");
 		$self->send_data('r', 'bye-bye');
 		$fh->close;
 		return;
+	    } elsif ($data[0] eq '__doit_rpc_ping') {
+		$d->(" got __doit_rpc_ping command");
+		$self->send_data('r', 'pong');
+		next;
 	    }
 	    $d->(" calling method $data[0]");
 	    my($rettype, @ret) = $self->runner->call_wrapped_method($context, @data);
@@ -2097,9 +2125,12 @@ use warnings;
 	    my($context, @data) = $self->receive_data;
 	    if (!defined $context) {
 		return;
-	    } elsif ($data[0] =~ m{^exit$}) {
+	    } elsif ($data[0] eq 'exit') {
 		$self->send_data('r', 'bye-bye');
 		return;
+	    } elsif ($data[0] eq '__doit_rpc_ping') {
+		$self->send_data('r', 'pong');
+		return;
 	    }
 	    open my $oldout, ">&STDOUT" or die $!;
 	    if (Doit::IS_WIN) {
@@ -2232,6 +2263,7 @@ use warnings;
 	my @cmd_worker =
 	    (
 	     'sudo', @sudo_opts, $perl, "-I".File::Basename::dirname(__FILE__), "-e",
+	     ($debug ? 'BEGIN { warn qq{DEBUG: sudo worker started...\n}}' : '') .
 	     Doit::_ScriptTools::self_require() .
 	     q{my $d = Doit->init; } .
 	     Doit::_ScriptTools::add_components(@components) .
@@ -2248,21 +2280,100 @@ use warnings;
 	    exec @cmd_worker;
 	    die "Failed to run '@cmd_worker': $!";
 	}
+	$self->{worker_pid} = $worker_pid;
 
 	# Run the client --- must also run under root for socket
 	# access.
 	my($in, $out);
-	my @cmd_comm = ('sudo', @sudo_opts, $perl, "-I".File::Basename::dirname(__FILE__), "-MDoit", "-e",
+	my @cmd_comm = (($LANS_PREFIX ? () : ('sudo', @sudo_opts)),
+	                $perl, "-I".File::Basename::dirname(__FILE__), "-MDoit", "-e",
 			q{Doit::Comm->comm_to_sock("} . $LANS_PREFIX . $sock_path . q{", debug => shift)}, !!$debug);
 	Doit::Log::trace("sudo comm: @cmd_comm") if DOIT_TRACE;
 	warn "comm perl cmd: @cmd_comm\n" if $debug;
 	my $comm_pid = IPC::Open2::open2($out, $in, @cmd_comm);
 	$self->{rpc} = Doit::RPC::Client->new($out, $in, label => "sudo:", debug => $debug);
+	$self->{comm_pid} = $comm_pid;
+
+	# set terminal again to "sane" settings, needed for sudo with use_pty configured
+	$self->{rpc}->wait_ready;
+	$self->_reset_terminal(debug => $debug);
 
 	$self;
     }
 
-    sub DESTROY { }
+    sub _reset_terminal {
+	my(undef, %opts) = @_;
+	my $debug = delete $opts{debug};
+	die "Unhandled options: " . join(" ", %opts) if %opts;
+
+	if (!-t STDIN) {
+	    if ($debug) {
+		info "Not a terminal, no reset needed";
+	    }
+	    return;
+	}
+
+	if (Doit::Util::is_in_path('stty')) {
+	    my @cmd = qw(stty sane);
+	    if ($debug) {
+		info "Reset terminal using @cmd...";
+	    }
+	    system(@cmd);
+	    if ($? == 0) {
+		return 1;
+	    }
+	    warning "_reset_terminal: '@cmd' failed, try POSIX.pm fallback...";
+	} else {
+	    warning "_reset_terminal: stty not available, try POSIX.pm fallback...";
+	}
+
+	if (!eval {
+	    require POSIX;
+
+	    my $term = POSIX::Termios->new();
+	    my $fd = fileno(STDIN);
+	    return 1 unless -t $fd;
+
+	    if ($debug) {
+		info "Reset terminal using POSIX.pm...";
+	    }
+
+	    $term->getattr($fd);
+
+	    my $lflag = $term->getlflag();
+	    my $iflag = $term->getiflag();
+	    my $oflag = $term->getoflag();
+
+	    $lflag |= POSIX::ICANON() | POSIX::ISIG() | POSIX::ECHO();
+
+	    $iflag |= POSIX::ICRNL();  # Convert CR to NL on input (fixes Enter key behavior)
+	    $oflag |= POSIX::OPOST();  # Enable output processing (fixes line endings)
+
+	    $term->setlflag($lflag);
+	    $term->setiflag($iflag);
+	    $term->setoflag($oflag);
+	    $term->setattr($fd, POSIX::TCSANOW());
+
+	    1;
+	}) {
+	    warning "Running _reset_terminal failed: $@";
+	    0;
+	} else {
+	    1;
+	}
+    }
+
+    sub DESTROY {
+	my $self = shift;
+	if ($self->{rpc}) {
+	    if ($self->{rpc}->{outfh}) {
+		$self->{rpc}->call_remote('exit');
+	    }
+	    kill TERM => $self->{comm_pid};
+	    $self->{rpc}->_reap_process($self->{comm_pid});
+	    $self->{rpc}->_reap_process($self->{worker_pid});
+	}
+    }
 
 }
 
@@ -2495,7 +2606,7 @@ use warnings;
 	my $d;
 	if ($debug) {
 	    $d = sub ($) {
-		Doit::Log::info("COMM: $_[0]");
+		Doit::Log::info("COMM: $_[0]\r");
 	    };
 	} else {
 	    $d = sub ($) { };
